!-------------------------------------------------------------------------------
!    This source code is part of
!    REIN (Replica-Exchange INterface program) version 1.0
!
!    Written by Naoyuki Miyashita and Yuji Sugita
!    Copyright (c) 2012, RIKEN, Japan
!
!    This program is free software; you can redistribute it and/or
!    modify it under the terms of the GNU General Public License
!    as published by the Free Software Foundation, either 
!    version 3 of the License, or (at your option) any later version.
!
!    This program is distributed in the hope that it will be useful,
!    but WITHOUT ANY WARRANTY; without even the implied 
!    warranty of MERCHANTABILITY or FITNESS FOR 
!    A PARTICULAR PURPOSE.  
!    See the GNU General Public License for more details.
!
!    You should have received a copy of the GNU General 
!    Public License along with REIN.  
!    If not, see <http://www.gnu.org/licenses/>.
!-------------------------------------------------------------------------------

!--------1---------2---------3---------4---------5---------6---------7---------8
!
!  Module : control_mod
!> @brief   REIN input control file modules
!! @author  Norio Takase (NT), Naoyuki Miyashita (NM)
!! @date    NM: 2009/04/02
!
!--------1---------2---------3---------4---------5---------6---------7---------8

module control_mod

  use common_mod
  use constants_mod

  implicit none
  private

  ! structures
  type, private :: s_key_value
    character(icha_2h) :: key
    character(icha_2h) :: equ
    character(icha_2h) :: value
    integer            :: num
  end type s_key_value

  type, private :: s_ctrl_keywords
    character(icha_2h) :: key_str                 = ''
    character(icha_2h) :: value_char(1:imax_axis) = ''  ! key_num = 1
    integer            :: value_int (1:imax_axis) = 0   ! key_num = 2
    double precision   :: value_real(1:imax_axis) = 0.0 ! key_num = 3
    integer            :: num_array
    integer            :: key_num
    integer            :: keyword_num
  end type s_ctrl_keywords

  ! parameter lists
  integer,                parameter :: max_key      = 39
  integer,                parameter :: max_key_char = 23
  integer,                parameter :: max_key_int  = 13
  integer,                parameter :: max_key_real =  3

  character(icha_2h),     parameter :: keyword_char_list(1:max_key_char) = (/ &
       'input_replica_rank     ', &
       'input_replica_values   ', &
       'input_replica_on       ', &
       'input_exchange_pattern ', &
       'input_axis_exchange    ', &
       'input_restart_tscale   ', &
       'output_restart_rank    ', &
       'output_restart_tscale  ', &
       'output_replica_rank    ', &
       'output_replica_energy  ', &
       'program_key            ', &
       'program_name           ', &
       'parallel_exec          ', &
       'parallel_module_md     ', &
       'option_md              ', &
       'extend_option_md       ', &
       'md_program_type        ', &
       'velocity_rescaling     ', &
       'initial_file_name      ', &
       'io_file_name           ', &
       'axis_type_name         ', &
       'parameter_setting      ', &
       'periodic               '/)

  character(icha_2h),     parameter :: keyword_int_list(1:max_key_int) = (/ &
       'first_step             ', &
       'equiv_steps            ', &
       'product_steps          ', &
       'number_of_axis         ', &
       'number_of_exchange     ', &
       'axis_index             ', &
       'number_of_replicas     ', &
       'axis_control_param     ', &
       'axis_num_of_param      ', &
       'axis_param             ', &
       'sleep_time             ', &
       'index_of_initial_parameter ', &
       'number_of_cpus_for_exec'/)
  character(icha_2h),     parameter :: keyword_real_list(1:max_key_real) = (/ &
       'base_temperature       ', &
       'periodic_value_min     ', &
       'periodic_value_max     '/)


  ! subroutines
  public  :: input_ctrl
  private :: input_getc
  private :: input_keyfind
  private :: input_keyfind2
  private :: input_assign

contains

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine  input_ctrl
  !> @brief      input control file
  !! @authors    NT, NM
  !! @date       2009/04/22 (NM)
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine input_ctrl()
    
    ! local variables
    type(s_key_value),         allocatable :: key_and_value(:)
    type(s_ctrl_keywords),     allocatable :: keywd(:)
    integer                                :: i_list, io_get
    character(icha_2h)                     :: argv


    ! allocate memory
    !
    allocate(key_and_value(1:imax))
    allocate(keywd(1:imax))


    ! read input file
    !
    call getarg(1, argv)

    io_get = 10 
    open(io_get, &
         file   = trim(adjustl(argv)), &
         status = 'old',       &
         form   = 'formatted', &
         err    = 90) 

    call input_getc(io_get, key_and_value, i_list)

    close(io_get)


    ! assign input paramters
    !
    call input_keyfind2(i_list, key_and_value, keywd)
    call input_assign  (keywd)


    ! deallocate memory
    deallocate(key_and_value)
    deallocate(keywd)

    return

90  write(io_out,*) 'Error: input file is not found. :'//trim(argv)
    write(io_out,*) ' '
    stop

  end subroutine input_ctrl


  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine  input_getc
  !> @brief      get the strings from argv (key and value)
  !! @authors    NT, NM
  !! @date       2009/04/22 (NM)
  !! @param[in]    io_get
  !! @param[inout] kv
  !! @param[out]   i_list
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine input_getc(io_get, kv, i_list)

    ! formal arguments
    integer,           intent(in)    :: io_get
    type(s_key_value), intent(inout) :: kv(:)
    integer,           intent(out)   :: i_list

    ! local variables
    integer          :: i, ist
    integer          :: sw_keep
    character(1)     :: ctmp
    character(80)    :: ktmp, keep


    i_list=1
    sw_keep = 0

    do i = 1, imax

      read(io_get,'(a80)',iostat=ist) ktmp

      if (ist == -1) then
        print *, "INPUT FILE eof=",ist
        exit
      end if

      ctmp = adjustl(ktmp)

      if ( (trim(adjustl(ctmp)) == "#") .or.  &
           (trim(adjustl(ctmp)) == "/") .or.  &
           (trim(adjustl(ctmp)) == "!") .or.  &
           (trim(adjustl(ctmp)) == "<") .or.  &
           (trim(adjustl(ctmp)) == ">") .or.  &
           (trim(adjustl(ctmp)) ==  "")) then 
        cycle
      else if (trim(adjustl(ctmp)) == "&") then
        sw_keep = 1
      else
        sw_keep = 0
      end if

      if (sw_keep == 1) then
        read(ktmp,*) ctmp, keep
        kv(i_list-1)%value = trim(adjustl(kv(i_list-1)%value))// &
                             trim(adjustl(keep))
      else
        read(ktmp,*) kv(i_list)%key,ctmp,kv(i_list)%value
        kv(i_list)%num = i_list
        i_list = i_list + 1
      end if
    end do
    i_list = i_list - 1

    do i = 1, i_list
      write(io_out,*) kv(i)%num, trim(kv(i)%key), " <=> ",trim(kv(i)%value) 
    end do

    write(io_out,*) ' '

    return

  end subroutine input_getc


  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine  input_keyfind
  !> @brief      get any values which you want
  !! @authors    NT, NM
  !! @date       2009/04/22 (NM)
  !! @param[in]  i_list
  !! @param[in]  kv
  !! @param[in]  key_1
  !! @param[out] i_out_1
  !! @param[out] out_1
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine input_keyfind(i_list, kv, key_1, i_out_1, out_1)

    ! formal arguments
    integer,            intent(in) :: i_list
    type(s_key_value),  intent(in) :: kv(:)
    character(icha_2h), intent(in) :: key_1
    integer,            intent(out):: i_out_1
    character(icha_2h), intent(out):: out_1(:)

    ! local variables
    integer             :: i


    i_out_1 = 0

    do i = 1, i_list
      if (kv(i)%key == key_1) then
        i_out_1 = i_out_1 + 1
        out_1(i_out_1) = trim(adjustl(kv(i)%value))
      end if
    end do

    return

  end subroutine input_keyfind


  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine  input_keyfind2
  !> @brief      modified the values
  !! @authors    NT, NM
  !! @date       2009/04/22 (NM)
  !! @param[in]  i_list
  !! @param[in]  kv
  !! @param[out] keywd
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine input_keyfind2(i_list, kv, keywd)

    ! formal arguments
    integer,                        intent(in)  :: i_list
    type(s_key_value),              intent(in)  :: kv(:)
    type(s_ctrl_keywords),          intent(out) :: keywd(:)

    ! local variables
    integer                         :: i_out_1
    integer                         :: i,j, i_key
    character(icha_2h)              :: key_string
    character(icha_16)              :: ni
    character(icha_2h), allocatable :: out_1(:)


    allocate(out_1(1:imax_axis))

    ! char
    i_key=0
    do i = 1, max_key_char

      i_key      =  i_key+1
      key_string =  keyword_char_list(i)

      call input_keyfind(i_list, kv, key_string, i_out_1, out_1)

      keywd(i_key)%key_str      = key_string
      keywd(i_key)%num_array    = i_out_1
      keywd(i_key)%key_num      = 1
      keywd(i_key)%keyword_num  = i_key

      do j = 1, i_out_1
        keywd(i_key)%value_char(j) = out_1(j)
      end do

    end do

    ! integer
    do i = 1, max_key_int

      i_key      =  i_key+1
      key_string =  keyword_int_list(i)

      call input_keyfind(i_list, kv, key_string, i_out_1, out_1)

      keywd(i_key)%key_str     = key_string
      keywd(i_key)%num_array   = i_out_1
      keywd(i_key)%key_num     = 2
      keywd(i_key)%keyword_num = i_key

      do j = 1, i_out_1
        ni = trim(adjustl(out_1(j)))
        ! trans
        read(ni,'(I8)') keywd(i_key)%value_int(j)
      end do
    end do

    ! real
    do i = 1, max_key_real

      i_key      =  i_key+1
      key_string =  keyword_real_list(i)

      call input_keyfind(i_list, kv, key_string, i_out_1, out_1)

      keywd(i_key)%key_str     = key_string
      keywd(i_key)%num_array   = i_out_1
      keywd(i_key)%key_num     = 3
      keywd(i_key)%keyword_num = i_key

      do j = 1, i_out_1
        ni = trim(adjustl(out_1(j)))
        !trans
        read(ni,'(F16.8)') keywd(i_key)%value_real(j)
      end do

    end do

    deallocate(out_1)

    return

  end subroutine input_keyfind2


  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Function    i_keyword
  !> @brief      get any values which you want
  !! @authors    NT, NM
  !! @date       2009/04/22 (NM)
  !! @param[in]  keywd
  !! @param[in]  inp_keywd
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  integer function i_keyword(keywd, inp_keywd)

    ! formal arguments
    type(s_ctrl_keywords), intent(in)  :: keywd(:)
    character(icha_2h),    intent(in)  :: inp_keywd

    ! local variables
    integer              :: i


    do i = 1, max_key
      if (trim(adjustl(keywd(i)%key_str)) == trim(adjustl(inp_keywd))) then
        i_keyword = keywd(i)%keyword_num
      end if
    end do

    return

  end function i_keyword


  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine  input_assign
  !> @brief      get any values which you want
  !!             If you make new input key, you should add it here.
  !! @authors    NT, NM
  !! @date       2009/04/22 (NM)
  !! @param[in]  keywd
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine input_assign(keywd)

    ! formal arguments
    type(s_ctrl_keywords),      intent(in)    :: keywd(:)

    ! local variables
    integer                     :: i, j, k, ii
    character(icha_2h)          :: str


    ! INPUT FILE SECTION
    !
    str                     = "input_replica_rank"
    Ginput_replica_rank     = keywd(i_keyword(keywd, str))%value_char(1)

    str                     = "input_replica_values"
    Ginput_replica_values   = keywd(i_keyword(keywd, str))%value_char(1)

    str                     = "input_replica_on"
    Ginput_replica_on       = keywd(i_keyword(keywd, str))%value_char(1)

    str                     = "input_exchange_pattern"
    Ginput_exchange_pattern = keywd(i_keyword(keywd, str))%value_char(1)

    str                     = "input_axis_exchange"
    Ginput_axis_exchange    = keywd(i_keyword(keywd, str))%value_char(1)

    str                     = "input_restart_tscale"
    Ginput_restart_tscale   = keywd(i_keyword(keywd, str))%value_char(1)

    ! OUTPUT FILE SECTION
    !
    str                     = "output_restart_rank"
    Goutput_restart_rank    = keywd(i_keyword(keywd, str))%value_char(1)

    str                     = "output_restart_tscale"
    Goutput_restart_tscale  = keywd(i_keyword(keywd, str))%value_char(1)

    str                     = "output_replica_rank"
    Goutput_replica_rank    = keywd(i_keyword(keywd, str))%value_char(1)

    str                     = "output_replica_energy"
    Goutput_replica_energy  = keywd(i_keyword(keywd, str))%value_char(1)

    ! MD PROGRAM SECTION
    !
    str                     = "program_name"             ! 13
    Gprogram_name           = keywd(i_keyword(keywd, str))%value_char(1)

    str                     = "parallel_exec"
    Gparallel_exec          = keywd(i_keyword(keywd, str))%value_char(1)

    str                     = "parallel_module_md"
    Gparallel_module_md     = keywd(i_keyword(keywd, str))%value_char(1)

    str                     = "option_md"
    Goption_md              = keywd(i_keyword(keywd, str))%value_char(1)

    str                     = "extend_option_md"
    Gextend_option_md       = keywd(i_keyword(keywd, str))%value_char(1)

    str                     = "velocity_rescaling"
    Gvelocity_rescaling     = keywd(i_keyword(keywd, str))%value_char(1)

    str                     = "base_temperature"         ! 1
    Gbase_temperature       = keywd(i_keyword(keywd, str))%value_real(1)

    ! EXCHANGE STEPS
    !
    str                     = "first_step"
    Gfirst_step             = keywd(i_keyword(keywd, str))%value_int(1)

    str                     = "equiv_steps"
    Gequiv_steps            = keywd(i_keyword(keywd, str))%value_int(1)

    str                     = "product_steps"
    Gproduct_steps          = keywd(i_keyword(keywd, str))%value_int(1)

    ! REPLICAS SECTION
    !
    str                     = "initial_file_name"
    Ginitial_file_name      = keywd(i_keyword(keywd, str))%value_char(1)

    str                     = "io_file_name"
    Gio_file_name           = keywd(i_keyword(keywd, str))%value_char(1)

    ! MULTI DIMENSION AXIS
    !
    str                     = "number_of_axis"           ! 32
    Gnumber_of_axis         = keywd(i_keyword(keywd, str))%value_int(1)

    str                     = "number_of_exchange"
    Gnumber_of_exchange     = keywd(i_keyword(keywd, str))%value_int(1)

    ! per AXIS
    str                     = "axis_index"               ! 1
    do i = 1, keywd(i_keyword(keywd, str))%num_array
      Gaxis_index(i)        = keywd(i_keyword(keywd, str))%value_int(i)
    end do

    str                     = "axis_type_name"           ! 1
    do i = 1, keywd(i_keyword(keywd, str))%num_array
      Gaxis_type_name(i)    = keywd(i_keyword(keywd, str))%value_char(i)
    end do

    str                     = "number_of_replicas"       ! 22
    do i = 1, keywd(i_keyword(keywd, str))%num_array
      Gnumber_of_replicas(i)= keywd(i_keyword(keywd, str))%value_int(i)
    end do

    str                     = "axis_control_param"       ! 1
    do i = 1, keywd(i_keyword(keywd, str))%num_array
      Gaxis_control_param(i)= keywd(i_keyword(keywd, str))%value_int(i)
    end do

    str                     = "axis_num_of_param"        ! 1
    do i = 1, keywd(i_keyword(keywd, str))%num_array
      Gaxis_num_of_param(i) = keywd(i_keyword(keywd, str))%value_int(i)
    end do

    str                     = "axis_param"               ! 1
    do j = 1, Gnumber_of_axis

      ii = 0

      if (j == 1) then
        ii = 0
      else
        do k = 1, Gnumber_of_axis - 1
          ii = ii + Gaxis_num_of_param(k)
        end do
      end if
      do i = 1, Gaxis_num_of_param(j)
        Gaxis_index_coef(i,j) &
             = keywd(i_keyword(keywd, str))%value_int(i + (j - 1)*ii)
      end do
    end do

    str                     = "periodic"                 ! 1
    do i = 1, keywd(i_keyword(keywd, str))%num_array
      Gperiodic(i)          = keywd(i_keyword(keywd, str))%value_char(i)
    end do

    str                     = "periodic_value_min"       ! 1
    do i = 1, keywd(i_keyword(keywd, str))%num_array
      Gperiodic_values(i,1) = keywd(i_keyword(keywd, str))%value_real(i)
    end do

    str                     = "periodic_value_max"       ! 1
    do i = 1, keywd(i_keyword(keywd, str))%num_array
      Gperiodic_values(i,2) = keywd(i_keyword(keywd, str))%value_real(i)
    end do

    ! MACHINE CONDITION SECTION
    !
    str                     = "sleep_time"               ! 1
    Gsleep_time             = keywd(i_keyword(keywd, str))%value_int(1)

    str                     = "number_of_cpus_for_exec"  ! 1
    Gnumber_of_cpus_for_exec= keywd(i_keyword(keywd, str))%value_int(1)

    return

  end subroutine input_assign

end module control_mod
