!-------------------------------------------------------------------------------
!    This source code is part of
!    REIN (Replica-Exchange INterface program) version 1.0
!
!    Written by Naoyuki Miyashita and Yuji Sugita
!    Copyright (c) 2012, RIKEN, Japan
!
!    This program is free software; you can redistribute it and/or
!    modify it under the terms of the GNU General Public License
!    as published by the Free Software Foundation, either 
!    version 3 of the License, or (at your option) any later version.
!
!    This program is distributed in the hope that it will be useful,
!    but WITHOUT ANY WARRANTY; without even the implied 
!    warranty of MERCHANTABILITY or FITNESS FOR 
!    A PARTICULAR PURPOSE.  
!    See the GNU General Public License for more details.
!
!    You should have received a copy of the GNU General 
!    Public License along with REIN.  
!    If not, see <http://www.gnu.org/licenses/>.
!-------------------------------------------------------------------------------

!--------1---------2---------3---------4---------5---------6---------7---------8
!
!  Module : fileio_mod
!> @brief   REIN file i/o modules
!! @author  Naoyuki Miyashita (NM)
!! @date    NM: 2009/04/02
!
!--------1---------2---------3---------4---------5---------6---------7---------8

module fileio_mod

  use common_mod
  use constants_mod
  
  implicit none
  private

  ! constants (IO)
  integer, public, parameter :: InvalidIO  = 1
  integer, public, parameter :: FirstIO    = 7
  integer, public, parameter :: MaxIOCount = 10000

  ! constants (File check)
  integer, public, parameter :: RetryCount = 100

  ! global variables
  logical                    :: busy_io(1:MaxIOCount) = .false.

  ! subroutines
  public  :: create_local_input
  public  :: compare_file_size
  public  :: compare_file_line
  public  :: check_file_size
  public  :: check_file_line
  public  :: check_file_tail
  public  :: get_io
  public  :: free_io
  public  :: make_filename
  public  :: add_directory

  private :: create_input_tagrec
  private :: read_line
  private :: print_line
  private :: scan_dquot
  private :: scan_dquot_ex
  private :: chk_file_exist

contains

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine  create_local_input
  !> @brief      create local input files
  !! @authors    NT, NM
  !! @date       2012/07/18 
  !! @param[in]  base_file    base file name
  !! @param[in]  target_file  target file name
  !! @param[in]  target_nline target number of lines
  !! @param[in]  local        md local information
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine create_local_input(base_file,    &
                                target_file,  &
                                target_nline, &
                                local)

    ! formal arguments
    character(icha_2h),         intent(in) :: base_file
    character(icha_2h),         intent(in) :: target_file
    integer,                    intent(in) :: target_nline
    type(s_local_info),         intent(in) :: local

    ! local variables
    integer                     ::  io_get, io_put


    call get_io(io_get)
    call get_io(io_put)

    call create_input_tagrec(io_get,      &
                             io_put,      &
                             base_file,   &
                             target_file, &
                             local)

    call free_io(io_put)
    call free_io(io_get)

    call compare_file_line(target_file, target_nline)

    return

  end subroutine create_local_input


  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine  compare_file_size
  !> @brief      compare file size
  !! @authors    NT, NM
  !! @date       2012/07/18 
  !! @param[in]  filename  file name
  !! @param[in]  nbyte_ref reference file size in byte
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine compare_file_size(filename, nbyte_ref)
    
    ! formal arguments
    character(icha_2h), intent(in) :: filename
    integer,            intent(in) :: nbyte_ref

    ! local variables
    integer             :: i, nbyte


    do i = 1, RetryCount
      call check_file_size(filename, nbyte)
      if (nbyte >= nbyte_ref) &
        exit
      write(io_out,'(A,A,A,I0,A,I0,A,I3)') &
           "Compare_File_Size> File:", trim(filename), &
           "  Tgt:", nbyte, "  Ref:", nbyte_ref, "  (Retry:", i
      call sleep(Gsleep_time)
    end do

    if (i >= RetryCount) then 
      write(io_out,'(a,a)') &
           'Compare_File_Size> ERROR: file is not completed :', &
           trim(filename)
      write(io_out,'(a)') '  ## You need larger "sleep_time" value'
      stop
    end if

    return

  end subroutine compare_file_size


  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine  compare_file_line
  !> @brief      compare file line
  !! @authors    NT, NM
  !! @date       2012/07/18 
  !! @param[in]  filename  file name
  !! @param[in]  nline_ref reference number of lines
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine compare_file_line(filename, nline_ref)

    ! formal arugments
    character(icha_2h), intent(in) :: filename
    integer,            intent(in) :: nline_ref

    ! local variables
    integer             :: i, nline


    do i = 1, RetryCount
      call check_file_line(filename, nline)
      if (nline >= nline_ref) &
        exit
      write(io_out,'(a,a,a,i0,a,i0,a,i3)') &
           'Compare_File_Line> File:', trim(filename), &
           '  Tgt:', nline, '  Ref:', nline_ref, '  (Retry:', i
      call sleep(Gsleep_time)
    end do

    if (i >= RetryCount) then 
      write(io_out,'(a,a)') &
           'Compare_File_Line> ERROR: file is not completed :', &
           trim(filename)
      write(io_out,'(a)') '  ## You need larger "sleep_time" value'
      stop
    end if

    return

  end subroutine compare_file_line


  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine  check_file_size
  !> @brief      check file size
  !! @authors    NT, NM
  !! @date       2012/07/18 
  !! @param[in]  filename  file name
  !! @param[out] nbyte     file size in byte
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine check_file_size(filename, nbyte)

    ! formal arguments
    character(icha_2h), intent(in)  :: filename
    integer,            intent(out) :: nbyte

    ! local variables
    integer             :: io_get, ierr


    nbyte = 0

    call get_io(io_get)
    open(io_get, &
         file   = trim(adjustl(filename)), &
         status = 'old', &
         iostat = ierr)

    if (ierr /= 0) then
       call free_io(io_get)
       return
    end if

#ifdef FUJI
    inquire(UNIT=io_get, FLEN=nbyte)
#else
    inquire(UNIT=io_get, SIZE=nbyte)  ! INTEL
#endif

    close(io_get)
    call free_io(io_get)

    return

  end subroutine check_file_size


  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine  check_file_line
  !> @brief      count file line
  !! @authors    NT, NM
  !! @date       2010/07/28 (NM)
  !! @param[in]  filename  file name
  !! @param[out] nline     number of lines
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine check_file_line(filename, nline)

    ! formal arguments
    character(icha_2h),    intent(in)  :: filename
    integer,               intent(out) :: nline

    ! local variables
    integer                :: io_get, ierr, ist
    character(icha_2h)     :: line


    nline = 0

    if (filename .eq. '') &
      return

    call get_io(io_get)
    open(io_get,               &
         file   = trim(adjustl(filename)), &
         status = "OLD",       &
         form   = 'formatted', &
         iostat = ierr)

    if (ierr /= 0) then
      call free_io(io_get)
      return
    end if

    io: do
      read(io_get,'(a)',iostat=ist) line

      select case(ist)
        case(0);
          nline = nline + 1;
        case default;
          exit
      end select
    end do io

    close(io_get)
    call free_io(io_get)

    return

  end subroutine check_file_line
   

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine  check_file_tail
  !> @brief      check file_tail
  !! @authors    NT, NM
  !! @date       2012/07/18 
  !! @note       fseek is not supported in  gfortran
  !! @param[in]    filename  file name
  !! @param[in]    string    string for search
  !! @param[in]    nbytes    number of backspace bytes
  !! @param[inout] found     flag for string was found or not
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine check_file_tail(filename, string, nbytes, found)

    ! formal arguments
    character(*),        intent(in)    :: filename
    character(*),        intent(in)    :: string
    integer,             intent(in)    :: nbytes
    integer,             intent(inout) :: found

    ! local variables
    integer              :: io_get, klen, ist
    character(icha_3h)   :: line


    found = 0
    klen  = LEN_TRIM(string)

    call get_io(io_get)
    open(io_get, file=filename, status='old', err=20)

    call fseek(io_get, -nbytes, 2)

    do while (.true.)
      read(io_get,'(a300)',iostat=ist) line
      if (ist /= 0) &
        exit

      if (line(1:klen) == string(1:klen)) then
        found = 1
        exit
      end if
    end do

    close(io_get)
20  call free_io(io_get)

    return

  end subroutine check_file_tail


  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine  get_io
  !> @brief      get random file unit number
  !! @authors    NT, NM
  !! @date       2012/07/09 
  !! @param[inout] io  i/o unut number
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine get_io(io)
    
    !$ use omp_lib
    ! parameter
    integer             :: MaxTry = 10

    ! formal arguments
    integer,            intent(inout) :: io

    ! local variables
    double precision    :: f
    integer             :: i
    integer             :: nn, nn2 ! HPC-K NYM

    ! recomendation of HPCI office for K
    ! !$omp critical

    ! ! O(0) search
    ! do i = 1, MaxTry
    
    !   call random_number(f)
    !   io = int(f * MaxIOCount)

    !   if (io < FirstIO) &
    !     io = FirstIO

    !   if (.not. busy_io(io)) then
    !    busy_io(io) = .true.
    !    exit
    !  end if

    !end do

    !! O(N) search
    !if (i > MaxTry) then

    nn = FirstIO   ! NYM
    !$ nn2 = (MaxIOCount -1)/ omp_get_num_threads() + 1
    !$ nn = FirstIO + nn2 * omp_get_thread_num()

      !do io = FirstIO, MaxIOCount
      do io = nn, MaxIOCount  ! NYM

        if (.not. busy_io(io)) then
          busy_io(io) = .true.
          exit
        end if

      end do

      if (io > MaxIOCount) then
        io = InvalidIO
        write(io_out,'(a)') 'Get_IO> ERROR: Free IO unit number is not found.'
      end if

    ! end if ! NYM

    ! NYM remove!$omp end critical

    return

  end subroutine get_io


  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine  free_io
  !> @brief      free random file unit number
  !! @authors    NT, NM
  !! @date       2012/07/09 
  !! @param[in] io  i/o unut number
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine free_io(io)

    ! formal arguments
    integer, intent(in) :: io


    ! NYM remove !$omp critical

    if (io > MaxIOCount .or. io == InvalidIO) then
      write(io_out,'(a)') &
           'Free_IO> ERROR: Invalid IO unit number.'
      stop
    end if

    if (.not. busy_io(io)) then
      write(io_out,'(a)') &
           'Free_IO> ERROR: Indicated IO unit number is not busy.'
    end if

    busy_io(io) = .false.

    ! NYM remove !$omp end critical

    return

  end subroutine free_io


  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine  make_filename
  !> @brief      make filename
  !! @authors    NT, NM
  !! @date       2012/06/26 
  !! @note       nrank    : main%name_pref_rank
  !!             dir_tag  : head_local_file
  !!             ext_tag  :   .inp  .colvar .crd ....  etc
  !!             n_steps  : main%name_i_steps
  !! @param[out] filename_out output file name
  !! @param[in]  dir_tag      directory tag
  !! @param[in]  ext_tag      file extension tag
  !! @param[in]  nrank        number of rank 
  !! @param[in]  n_steps      number of steps
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine make_filename(filename_out, &
                           dir_tag, &
                           ext_tag, &
                           nrank,   &
                           n_steps)

    ! formal arguments
    character(icha_2h), intent(out)   :: filename_out 
    character(icha_2h), intent(in)    :: dir_tag
    character(icha_2h), intent(in)    :: ext_tag
    character(icha_8 ), intent(in)    :: nrank
    character(icha_8 ), intent(in)    :: n_steps 


    filename_out = trim(adjustl(dir_tag)) &
    !            // "id."                  &      ! NYM
                // trim(adjustl(nrank))   &
                // "."                    &
                // trim(adjustl(n_steps)) &
                // trim(adjustl(ext_tag))

    return

  end subroutine make_filename


  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine  add_directory
  !> @brief      add directory
  !! @authors    NT, NM
  !! @date       2012/06/26 
  !! @param[out] filename_out output file name
  !! @param[in]  nrank        number of rank 
  !! @param[in]  dir_tag      directory tag
  !! @param[in]  filename_in  input file name
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine add_directory(filename_out, &
                           nrank,        &
                           dir_tag,      &
                           filename_in)

    ! formal arguments
    character(icha_2h), intent(out)  :: filename_out 
    character(icha_8),  intent(in)   :: nrank
    character(icha_2h), intent(in)   :: dir_tag
    character(icha_2h), intent(in)   :: filename_in


    filename_out = trim(adjustl(dir_tag)) &
                // trim(adjustl(nrank))   &
                // "/"                    &
                // trim(adjustl(filename_in))

    return

  end subroutine add_directory


  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine  create_input_tagrec
  !> @brief      CREATE INPUT FILE for each binary program using TAG
  !! @authors    NT, NM
  !! @date       2011/01/10 (NM)
  !! @note       READ from base_file, translation about TAG, WRITE to out_file
  !! @param[in]  io_get      i/o unit number
  !! @param[in]  io_put      i/o unit number
  !! @param[in]  base_file   base file name
  !! @param[in]  target_file base file name
  !! @param[in]  local       md local information
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine create_input_tagrec(io_get,      &
                                 io_put,      &
                                 base_file,   &
                                 target_file, &
                                 local)

    ! formal arguments
    integer,               intent(in) :: io_get
    integer,               intent(in) :: io_put
    character(icha_2h),    intent(in) :: base_file
    character(icha_2h),    intent(in) :: target_file
    type(s_local_info),    intent(in) :: local
                            
    ! local variables
    character(icha_2h)     :: line, line2
    integer                :: num_axis, num_coef, iquot
    integer                :: ist

    integer                :: ierr
    character(icha_16)     :: nn16

    character(icha_2h)     :: rein_indication
    character(icha_2h)     :: inp_command, inp_command2
    integer                :: iquot1, iquot2


    !check file
    !
    open(io_get, &
         file   = trim(adjustl(base_file)), &
         status = "OLD",       &
         form   = 'formatted', &
         iostat = ierr)

    open(io_put, &
         file   = trim(adjustl(target_file)), &
         status = "unknown",   &
         form   = 'formatted', &
         iostat=ierr)


    inp:do

      call read_line(io_get, line, ist)

      select case(ist)
      case(0);
      case default;
        exit
      end select

      line2 = trim(adjustl(line))
      iquot = scan(line2, ':')
      rein_indication = line2(1:iquot)

      select case (trim(adjustl(rein_indication)))
      case ("rein_inpfile:")
        call scan_dquot(line2, inp_command)
        line = trim(adjustl(inp_command)) // & 
             " " // trim(adjustl(local%local_input_crd))

      case ("rein_outfile:")
        call scan_dquot(line2, inp_command)
        line = trim(adjustl(inp_command)) // &
             " " // trim(adjustl(local%local_output_crd))

      case ("rein_trjfile:")
        call scan_dquot(line2, inp_command)
        line = trim(adjustl(inp_command)) // &
             " " // trim(adjustl(local%local_output_trj))

     case ("rein_propfile:")
        call scan_dquot(line2, inp_command)
        line = trim(adjustl(inp_command)) // &
             " " // trim(adjustl(local%local_output_prop))

      case ("rein_baseT:")
        call scan_dquot(line2, inp_command)
        write(nn16,'(f16.8)') Gbase_temperature
        line = trim(adjustl(inp_command)) // " " // trim(adjustl(nn16))

      case ("rein_control:")
        read(line2(iquot+1:200),*) num_axis, num_coef
        call scan_dquot(line2, inp_command)
        write(nn16,'(f16.8)') local%condition_values(num_coef)
        line = trim(adjustl(inp_command)) // " " // trim(adjustl(nn16))

      case ("rein_colvar:")
        call scan_dquot_ex(line2, inp_command, iquot1, iquot2)
        call scan_dquot_ex(line2(iquot2+2:icha_2h), inp_command2, iquot1,iquot2)
        line = trim(adjustl(inp_command)) &
             // " " // trim(adjustl(local%local_extra_file2))

      case default
      end select

      call print_line(io_put,line)
    end do inp

    close(io_get)
    close(io_put)

    return

  end subroutine create_input_tagrec

 
  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine  read_line
  !> @brief      read line
  !! @authors    NT, NM
  !! @date       2012/06/26 
  !! @param[in]  io   i/o unit number
  !! @param[out] line line
  !! @param[out] istout
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine read_line(io, line, istout)

    ! formal arguments
    integer,            intent(in)  :: io
    character(icha_2h), intent(out) :: line
    integer,            intent(out) :: istout

    ! local variables
    integer             :: ist


    read(io,'(a)',iostat=ist) line
    istout = ist

    return

  end subroutine read_line


  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine  print_line
  !> @brief      REIN file line printer
  !! @authors    NT, NM
  !! @date       2010/01/21 (NM)
  !! @param[in]  io  i/o unit number
  !! @param[in]  line line
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine print_line(io, line)

    ! formal arguments
    integer,            intent(in) :: io
    character(icha_2h), intent(in) :: line

    ! local variables
    integer             :: ist
    character(icha_8)   :: nn


    ist = len_trim(line)

    select case(ist)
    case (0)
      ist = 1

    case default

    end select

    write(nn,'(i8)') ist
    write(io,'(a)') line(1:ist)

    return

  end subroutine print_line


  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine  scan_dquot
  !> @brief      scannning the string in double quotation
  !! @authors    NT, NM
  !! @date       2010/07/14 (NM)
  !! @param[in]  inpline input line
  !! @param[out] outline output line
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine scan_dquot(inpline, outline)

    ! formal arguments
    character(icha_2h), intent(in) :: inpline
    character(icha_2h), intent(out):: outline

    ! local variables
    integer :: iquot1, iquot2


    iquot1  = scan(inpline,'\"')
    iquot2  = scan(inpline,'\"',.true.)
    iquot1  = iquot1+1
    iquot2  = iquot2-1
    outline = inpline(iquot1:iquot2)

    return

  end subroutine scan_dquot


  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine  scan_dquot_ex
  !> @brief      scannning the string in double quotation extended
  !! @authors    NT, NM
  !! @date       2010/07/14 (NM)
  !! @param[in]  inpline input line
  !! @param[out] outline output line
  !! @param[out] iquot1 left quotation
  !! @param[out] iquot2 right quotation
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine scan_dquot_ex(inpline, outline, iquot1, iquot2)

    ! formal arguments
    character(icha_2h), intent(in)  :: inpline
    character(icha_2h), intent(out) :: outline
    integer,            intent(out) :: iquot1
    integer,            intent(out) :: iquot2

    ! local variables
    integer             :: iq


    iquot1 = 0; iquot2=0
    iquot1  = scan(inpline,'\"')
    iquot1  = iquot1+1
    iq      = iquot1  +1
    iquot2  = scan(inpline(iq:icha_2h),'\"')
    iquot2  = iquot2 + iq -2
    outline = inpline(iquot1:iquot2)

    return

  end subroutine scan_dquot_ex
   

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Function    check_file_exist
  !> @brief      check file existence
  !! @authors    NT, NM
  !! @date       2010/07/28 (NM)
  !! @param[in] filename file name
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  integer function chk_file_exist(filename)

    ! formal arguments
    character(icha_2h), intent(in) :: filename

    ! local variables
    logical             :: lex


    !lex=.false.
    chk_file_exist=1
    inquire(file  = trim(adjustl(filename)), &
            exist = lex)
    !if(lex .eq. .false. ) chk_file_exist=1
    !if(lex .eq. .true.  ) chk_file_exist=0
    if (lex) &
      chk_file_exist = 0

    return

  end function chk_file_exist


end module fileio_mod
