!-------------------------------------------------------------------------------
!    This source code is part of
!    REIN (Replica-Exchange INterface program) version 1.0
!
!    Written by Naoyuki Miyashita and Yuji Sugita
!    Copyright (c) 2012, RIKEN, Japan
!
!    This program is free software; you can redistribute it and/or
!    modify it under the terms of the GNU General Public License
!    as published by the Free Software Foundation, either 
!    version 3 of the License, or (at your option) any later version.
!
!    This program is distributed in the hope that it will be useful,
!    but WITHOUT ANY WARRANTY; without even the implied 
!    warranty of MERCHANTABILITY or FITNESS FOR 
!    A PARTICULAR PURPOSE.  
!    See the GNU General Public License for more details.
!
!    You should have received a copy of the GNU General 
!    Public License along with REIN.  
!    If not, see <http://www.gnu.org/licenses/>.
!-------------------------------------------------------------------------------

!--------1---------2---------3---------4---------5---------6---------7---------8
!
!  Module : namd_module_mod
!> @brief   REIN NAMD modules
!! @author  Naoyuki Miyashita (NM)
!! @date    NM: 2009/07/02
!
!--------1---------2---------3---------4---------5---------6---------7---------8

module namd_module_mod

  use namd_rst_rescaling_mod
  use fileio_mod
  use parallel_mod
  use common_mod
  use constants_mod
  use mpi

  implicit none
  private
   
  ! subroutines
  public  :: namd_make_input
  public  :: namd_rescaling
  public  :: namd_watchdog
  public  :: namd_get_output
  private :: get_io_name
  private :: get_mdout
  private :: get_colvar
  private :: check_vel_endian

contains

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine  namd_make_input
  !> @brief      make NAMD input files
  !! @authors    NT, NM
  !! @date       2012/07/02 
  !! @param[in]    world  MPI information
  !! @param[in]    main   main information
  !! @param[inout] namd   namd local information
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine namd_make_input(world, main, namd)

    ! formal arguments
    type(s_comm_x),      intent(in)    :: world
    type(s_main_info),   intent(in)    :: main
    type(s_local_info),  intent(inout) :: namd

    ! local variables
    character(icha_2h)   :: str


    call get_io_name(namd)

    if (main%i_steps == Gfirst_step) then

      ! check numbre of input lines
      call check_file_line(namd%base_file, &
                 namd%local_nline_inputfile)

      if (main%flag_has_harmonic > 0) &
        call check_file_line(namd%local_extra_file1, &
                   namd%local_nline_extrafile1)

      ! check velocity file size, byte order
      str = trim(namd%local_dir) //'/'// &
            trim(namd%local_pre_input_file) //&
            ".vel"

      call check_file_size(str, namd%local_nbyte_velfile)
      call check_vel_endian(str)

    end if

    ! make input file
    !
    call create_local_input(                 &
                 namd%base_file,             &
                 namd%local_input_path,      &
                 namd%local_nline_inputfile, &
                 namd)

    if (main%flag_has_harmonic > 0) &
      call create_local_input(                &
                 namd%local_extra_file1,      &
                 namd%local_extra_path2,      &
                 namd%local_nline_extrafile1, &
                 namd)

    return
    
  end subroutine namd_make_input
   

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine  namd_rescaling
  !> @brief      rescaling NAMD velocity/coordinate files
  !! @authors    NT, NM
  !! @date       2009/12/08 (NM)
  !! @param[in]    world     MPI information
  !! @param[in]    main      main information
  !! @param[inout] namd      namd local information
  !! @param[in]    ireplica  replica number
  !! @param[in]    flag
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine namd_rescaling(world, main, namd, ireplica)

#ifdef FUJI
    use service_routines, only: rename
#endif
#ifdef INTEL
    use iflport
    use iflposix
    use ifport
#endif

    ! formal arguments
    type(s_comm_x),     intent(in)    :: world
    type(s_main_info),  intent(in)    :: main
    type(s_local_info), intent(inout) :: namd
    integer,            intent(in)    :: ireplica

    ! local variables
    double precision    :: Tpre, T, coef
    integer             :: i_status
    integer             :: it
    character(icha_2h)  :: tmp_file, tmp_file2
    character(icha_2h)  :: flag3


    it = main%flag_velocity_rescaling

    if (it == 0) &
      return

    if (main%i_steps == main%initial_steps) then
      coef = main%replica_tscale(ireplica,1)
      T    = main%replica_values(main%replica_rank(ireplica,1), it)
      Tpre = T/coef
    else
      Tpre = main%replica_values(main%replica_prerank(ireplica,1), it)
      T    = main%replica_values(main%replica_rank   (ireplica,1), it)
      coef = dsqrt(T / Tpre)
    end if

    if (dabs(coef-1.d0) > err16) then

#ifdef DEBUG
      write(io_out,*)'CHK:RESCALING final=', Tpre, T, coef
#endif

      tmp_file = trim(adjustl(namd%local_dir)) // "/" // &
                 trim(adjustl(namd%local_pre_input_file)) // &
                 ".vel"


      ! rename file
      !

      ! check file i/o complete
      call compare_file_size(tmp_file, namd%local_nbyte_velfile)

      tmp_file2 = trim(adjustl(tmp_file)) // ".org"
      i_status  = rename(trim(adjustl(tmp_file)), trim(adjustl(tmp_file2)))
      if (i_status /= 0) then
        write(io_out,*) '  Namd_Rescaling> ERROR. rename was failed. "', &
             trim(tmp_file), '" -> "', &
             trim(tmp_file2), '"'
        stop
      end if


      ! rescaling file
      !

      ! check file i/o complete
      call compare_file_size(tmp_file2, namd%local_nbyte_velfile)

      select case (trim(adjustl(Gvelocity_rescaling)))
      case ('v_namd')
        ! ASCII FILE RESCALING 
        flag3="ascii"

      case ('v_namd_bin')
        ! BINARY FILE RESCALING
        flag3="bin"

      case default
        write(6,*) '  Namd_Rescaling> WARNING. unknown type:', &
                   trim(Gvelocity_rescaling)

      end select

      call namd_rst_rescaling(coef, tmp_file2, tmp_file, flag3)

      ! check file i/o complete (final)
      call compare_file_size(tmp_file, namd%local_nbyte_velfile)

    end if

    return

  end subroutine namd_rescaling


  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine  namd_watchdog
  !> @brief      watchdog NAMD process
  !! @authors    NT, NM
  !! @date       2012/07/24 
  !! @param[in]    world      MPI information
  !! @param[in]    main       main information
  !! @param[inout] namd       namd local information
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine namd_watchdog(world, main, namd)

    ! parameter
    integer, parameter   :: WatchInterval = 1    ! 1 second
    integer, parameter   :: MaxWatch      = 7200 ! 2 hour

    ! formal arguments
    type(s_comm_x),      intent(in)    :: world
    type(s_main_info),   intent(in)    :: main
    type(s_local_info),  intent(inout) :: namd

    ! local variables
    integer              :: i, found, nline
    character(icha_2h)   :: fileout, filecol


    fileout = adjustl(namd%local_output_path)
    filecol = adjustl(namd%local_extra_path3)

    do i = 1, MaxWatch

      if (main%flag_has_harmonic > 0) then

        call check_file_tail(fileout, 'End of program', 100, found)
        call check_file_line(filecol, nline)
        if (found /= 0 .and. nline >= 3) &
          exit

      else

        call check_file_tail(fileout, 'End of program', 100, found)
        if (found /= 0) &
          exit

      end if

      call sleep(WatchInterval)

    end do

    if (i >= MaxWatch) then
      write(io_out,'(a)') &
           'Namd_WatchDog> Over maximum watch. NAMD is hunging-up?'
    end if

    return

  end subroutine namd_watchdog


  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine  namd_get_output
  !> @brief      get output for NAMD
  !! @authors    NT, NM
  !! @date       2012/07/09 
  !! @param[inout] main       main information
  !! @param[in]    namd       namd local information
  !! @param[in]    ireplica   replica number
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine namd_get_output(main, namd, ireplica)

    ! formal arguments
    type(s_main_info),  intent(inout) :: main
    type(s_local_info), intent(in)    :: namd
    integer,            intent(in)    :: ireplica

    ! local variables
    integer             :: j


    call get_mdout(main, namd)

    if (namd%num_axis_type /= 0) &
      call get_colvar(main, namd)

    return

  end subroutine namd_get_output


  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine  get_io_name
  !> @brief      get i/o name
  !! @authors    NT, NM
  !! @date       2010/07/14 (NM)
  !! @param[inout] namd  namd local information
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine get_io_name(namd)

    ! formal arguments
    type(s_local_info),   intent(inout) :: namd

    ! local variables
    character(icha_2h)    :: blank_cha


    ! file
    blank_cha = " "
    call make_filename(namd%local_file,     &
                       Gio_file_name,       &
                       blank_cha,           &
                       namd%name_pref_rank, &
                       namd%name_i_steps)

    ! file.inp
    blank_cha = ".inp"
    call make_filename(namd%local_input_file, &
                       Gio_file_name,         &
                       blank_cha,             &
                       namd%name_pref_rank,   &
                       namd%name_i_steps)

    !-------- run shell   for spawn command=xxxx.sh
    blank_cha = ".sh"
    call make_filename(namd%local_command_file, &
                       Gio_file_name,           &
                       blank_cha,               &
                       namd%name_pref_rank,     &
                       namd%name_i_steps)

    ! prefile
    blank_cha = " "
    call make_filename(namd%local_pre_input_file, &
                       Gio_file_name,             &
                       blank_cha,                 &
                       namd%name_pref_rank,       &
                       namd%name_ipre_steps)

    ! file.out
    blank_cha = ".out"
    call make_filename(namd%local_output_file, &
                       Gio_file_name,          &
                       blank_cha,              &
                       namd%name_pref_rank,    &
                       namd%name_i_steps)

    ! dir/file
    call add_directory(namd%local_path,     &
                       namd%name_pref_rank, &
                       Gio_file_name,       &
                       namd%local_file)

    ! dir/file.inp
    call add_directory(namd%local_input_path, &
                       namd%name_pref_rank,   &
                       Gio_file_name,         &
                       namd%local_input_file)

    ! dir/prefile
    call add_directory(namd%local_pre_input_path, &
                       namd%name_pref_rank,       &
                       Gio_file_name,             &
                       namd%local_pre_input_file)

    ! dir/file.out
    call add_directory(namd%local_output_path, &
                       namd%name_pref_rank,    &
                       Gio_file_name,          &
                       namd%local_output_file)

    namd%base_file  = trim(adjustl(Ginitial_file_name)) // ".inp"

    namd%local_dir =  trim(adjustl(Gio_file_name)) // &
                      trim(adjustl(namd%name_pref_rank))

    namd%local_input_crd  = &
                      trim(adjustl(namd%local_pre_input_file))
    namd%local_output_crd = &
                      trim(adjustl(namd%local_file))
    namd%local_output_trj = &
                      trim(adjustl(namd%local_file)) // ".dcd"

    namd%local_command_path = &
                      trim(adjustl(namd%local_dir)) // "/" // &
                      trim(adjustl(namd%local_command_file))

    ! colvar file
    namd%local_extra_file1 = &
                      trim(adjustl(Ginitial_file_name)) // ".colvar"
    namd%local_extra_path1 = &
                      trim(adjustl(namd%local_dir))  // "/" // &
                      trim(adjustl(namd%local_extra_file1))
    namd%local_extra_file2 = &
                      trim(adjustl(namd%local_file)) // ".colvar"
    namd%local_extra_path2 = &
                      trim(adjustl(namd%local_dir))  // "/" // &
                      trim(adjustl(namd%local_extra_file2))
    namd%local_extra_file3 = &
                      trim(adjustl(namd%local_file)) // &
                      ".colvars.traj"
    namd%local_extra_path3 = &
                      trim(adjustl(namd%local_dir)) // "/" // &
                      trim(adjustl(namd%local_extra_file3))

    ! command file
    namd%local_command = &
                      trim(adjustl(Gparallel_module_md))   // " " // &
                      trim(adjustl(Gparallel_exec))        // " " // &
                      trim(adjustl(Goption_md))            // " " // &
                      trim(adjustl(namd%local_input_path)) // " " // &
                      trim(adjustl(Gextend_option_md))     // " > " // &
                      trim(adjustl(namd%local_output_path))

    !write(6,*) '## base_file             :', TRIM(namd%base_file)
    !write(6,*) '##'
    !write(6,*) '## local_dir             :', TRIM(namd%local_dir)
    !write(6,*) '## local_file            :', TRIM(namd%local_file)
    !write(6,*) '## local_path            :', TRIM(namd%local_path)
    !write(6,*) '## local_input_file      :', TRIM(namd%local_input_file)
    !write(6,*) '## local_input_path      :', TRIM(namd%local_input_path)
    !write(6,*) '## local_output_file     :', TRIM(namd%local_output_file)
    !write(6,*) '## local_output_path     :', TRIM(namd%local_output_path)
    !write(6,*) '## local_pre_input_file  :', TRIM(namd%local_pre_input_file)
    !write(6,*) '## local_pre_input_path  :', TRIM(namd%local_pre_input_path)
    !write(6,*) '## local_command_file    :', TRIM(namd%local_command_file)
    !write(6,*) '## local_command_path    :', TRIM(namd%local_command_path)
    !write(6,*) '## local_extra_file1     :', TRIM(namd%local_extra_file1)
    !write(6,*) '## local_extra_path1     :', TRIM(namd%local_extra_path1)
    !write(6,*) '## local_extra_file2     :', TRIM(namd%local_extra_file2)
    !write(6,*) '## local_extra_path2     :', TRIM(namd%local_extra_path2)
    !write(6,*) '## local_extra_file3     :', TRIM(namd%local_extra_file3)
    !write(6,*) '## local_extra_path3     :', TRIM(namd%local_extra_path3)
    !write(6,*) '##'
    !write(6,*) '## local_input_crd       :', TRIM(namd%local_input_crd)
    !write(6,*) '## local_output_crd      :', TRIM(namd%local_output_crd)
    !write(6,*) '## local_output_trj      :', TRIM(namd%local_output_trj)
    !write(6,*) '##'
    !write(6,*) '## local_command         :', TRIM(namd%local_command)
    !write(6,*) '##'
    !write(6,*) '## name_pref_rank        :', TRIM(namd%name_pref_rank)
    !write(6,*) '## name_i_steps          :', TRIM(namd%name_i_steps)
    !write(6,*) '## name_ipre_steps       :', TRIM(namd%name_ipre_steps)

    return

  end subroutine get_io_name


  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine  get_mdout
  !> @brief      get namd output
  !! @authors    NT, NM
  !! @date       2012/07/18 
  !! @param[inout] main  main information
  !! @param[in]    namd  namd local information
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine get_mdout(main, namd)

    ! formal arguments
    type(s_main_info),  intent(inout) :: main
    type(s_local_info), intent(in)    :: namd

    ! local variables
    double precision    :: a(1:18)
    integer             :: io, ierr, i, stp
    character(icha_3h)  :: line, eline
    character(7)        :: head7


    ! Read energy line
    !
    eline = ''

    call get_io(io)
    open(io, &
         file   = adjustl(namd%local_output_path), &
         status = "OLD",       &
         form   = 'formatted', &
         iostat = ierr)

    do while (.true.)

      read(io,'(a300)', iostat=ierr) line

      if (ierr /= 0) &
        exit

      if (line(1:5) .eq. 'ENERG') then
         eline = line
      end if

    end do

    close(io)
    call free_io(io)


    ! parse energy
    !
    read(eline, "(a7,I8,4F15.4,F20.4,4F15.4,F20.4,4F15.4,F20.4,4F15.4)") &
         head7, stp, (a(i),i=1,18,1)

    main%replica_extra_pote (namd%nrank) = a(8)
    main%replica_kinetice   (namd%nrank) = a(9)
    main%replica_energy     (namd%nrank) = a(10)
    main%replica_temperature(namd%nrank) = a(11)
    main%replica_potentiale (namd%nrank) = a(12)

    return

20  write(io_out,'(a)') 'Get_Mdout> NAMD output read error.'

  end subroutine get_mdout


  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine  get_colvar
  !> @brief      get output colvar
  !! @authors    NT, NM
  !! @date       2012/07/18 
  !! @param[inout] main       main information
  !! @param[in]    namd       namd local information
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine get_colvar(main, namd)

    ! parameters
    integer,            parameter :: max_limitter = 10000

    ! formal arguments
    type(s_main_info),  intent(inout) :: main
    type(s_local_info), intent(in)    :: namd

    ! local variables
    double precision    :: tmp_stock(1:10)
    integer             :: io, i, j, ierr, ist
    integer             :: num
    character(icha_2h)  :: line


    call compare_file_line(adjustl(namd%local_extra_path3), 3)

    call get_io(io)
    open(io, &
         file   = adjustl(namd%local_extra_path3), &
         status = "OLD",       &
         form   = 'formatted', &
         iostat = ierr)
    read(io,'(a200)') line 

    do j = 1, max_limitter
      read(io,*,iostat=ist) num, &
           (tmp_stock(i), i = 1, namd%num_axis_type, 1)
      if (ist /= 0) &
        exit
    end do

    close (io)
    call free_io(io)

    do j =1, namd%num_axis_type
      main%replica_param_p(namd%nrank, j) = tmp_stock(j)
    end do

    return

  end subroutine get_colvar


  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine  check_vel_endian
  !> @brief      check and correct byte order
  !! @authors    NT
  !! @date       2012/07/26 
  !! @param[in]  filename  NAMD velocity file nam
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine check_vel_endian(velfile)

    ! formal arguments
    character(*),     intent(in)  :: velfile

    ! local variables
    integer           :: io, natom, nbyte

    double precision, allocatable :: crd(:)


    call get_io(io)
#ifdef GFORT
    open(io,                        &
         file   = adjustl(velfile), &
         status = 'old',            &
         form   = 'unformatted',    &
         access = 'stream')
#else
    open(io,                        &
         file   = adjustl(velfile), &
         status = 'old',            &
         form   = 'binary',         &
         access = 'sequential')
#endif

    read(io) natom

#ifdef FUJI
    inquire(UNIT=io, FLEN=nbyte)
#else
    inquire(UNIT=io, SIZE=nbyte)  ! INTEL
#endif

    close(io)

    if (nbyte /= 4 + 24 * natom) then

      open(io, &
           file    = adjustl(velfile),&
#ifdef FUJI
           convert = 'little_endian', &
#else
           convert = 'big_endian',    &
#endif
           status  = 'old',           &
#ifdef GFORT
           form    = 'unformatted',   &
           access  = 'stream')
#else
           form    = 'binary',        &
           access  = 'sequential')
#endif

      read(io) natom

      allocate(crd(natom * 3))
      rewind(io)
      read(io) natom, crd
      close(io)
      
#ifdef GFORT
      open(io,                        &
           file   = adjustl(velfile), &
           status = 'replace',        &
           form   = 'unformatted',    &
           access = 'stream')
#else 
      open(io,                        &
           file   = adjustl(velfile), &
           status = 'replace',        &
           form   = 'binary',         &
           access = 'sequential')
#endif

      write(io) natom, crd
      close(io)

      write(io_out,'(a)') &
            'Check_Vel_Endian> Warning. NAMD velocity byte order was correctd.'
      write(io_out,'(a,a)') &
            '                >   File : ', trim(velfile)
#ifdef GFORT
      write(io_out,'(a27,i8,a10,f8.3)') &
            '                >   Natom: ', natom, '  Crd1,X: ', crd(1)
#else
      write(io_out,'(a,i,a,f8.3)') &
            '                >   Natom: ', natom, '  Crd1,X: ', crd(1)
#endif

      deallocate(crd)

    end if

    return

  end subroutine check_vel_endian

end module namd_module_mod
