!-------------------------------------------------------------------------------
!    This source code is part of
!    REIN (Replica-Exchange INterface program) version 1.0
!
!    Written by Naoyuki Miyashita and Yuji Sugita
!    Copyright (c) 2012, RIKEN, Japan
!
!    This program is free software; you can redistribute it and/or
!    modify it under the terms of the GNU General Public License
!    as published by the Free Software Foundation, either 
!    version 3 of the License, or (at your option) any later version.
!
!    This program is distributed in the hope that it will be useful,
!    but WITHOUT ANY WARRANTY; without even the implied 
!    warranty of MERCHANTABILITY or FITNESS FOR 
!    A PARTICULAR PURPOSE.  
!    See the GNU General Public License for more details.
!
!    You should have received a copy of the GNU General 
!    Public License along with REIN.  
!    If not, see <http://www.gnu.org/licenses/>.
!-------------------------------------------------------------------------------

!--------1---------2---------3---------4---------5---------6---------7---------8
! 
!> Program  REIN
!! @brief   Replica Exchange INterface (REIN)
!! @authors Naoyuki Miyashita (NM), Yuji Sugita (YS)
!! @date    NM: 2014/02/23 ver 1.0.4
!! @date    NM: 2012/02/18 ver 0.9
!! @date    NM: 2009/06/18 ver 0.1
!
!--------1---------2---------3---------4---------5---------6---------7---------8

program REIN

  use driver_mod
  use setup_mod
  use control_mod
  use common_mod
  use parallel_mod
  use constants_mod
  use omp_lib
  use mpi

  implicit none

  ! local variables
  type(s_comm_x)    :: world
  type(s_main_info) :: main
  double precision  :: t(5)
  double precision  :: dt(5)
  integer           :: ierr, iprovide


  ! Initialize MPI
  !

  t(1) = mpi_wtime()

  world%comm = MPI_COMM_WORLD

  call MPI_INIT_THREAD    (MPI_THREAD_MULTIPLE, iprovide, ierr)
  call MPI_COMM_RANK(world%comm, world%rank,   ierr)
  call MPI_COMM_SIZE(world%comm, world%nprocs, ierr)

  if (world%rank == 0) then
    write(io_out,*)'===================================================>'
    write(io_out,*)'=        REIN: Replica-Exchange INterface        ==>'
    write(io_out,*)'=     Naoyuki Miyashita (NYN) and Yuji Sugita    ==>'
    write(io_out,*)'=          date: 2012/11/02  ver.1.0.2           ==>'
    write(io_out,*)'===================================================>'
    write(io_out,*) ' '
    write(io_out,*)'number of process: ', world%nprocs
    write(io_out,*) ' '
  end if


  ! Setting and reading input file
  !

  call init_values

  if (world%rank == 0) then
    write(io_out,*) 'INPUT CTRL>  ======================================>'
    write(io_out,*) ' '
    call input_ctrl
  end if

  t(2) = MPI_WTIME()
  dt(1) = t(2) - t(1)


  ! Data bcasting in the main MPI
  !

  if (world%rank == 0) then
    write(io_out,*) 'INPUT BCASTING>  ==================================>'
    write(io_out,*) ' '
  end if

  call bcast_values(world)

  t(3) = mpi_wtime()
  dt(2) = t(3) - t(2)


  ! Reading and setting condition
  !

  if (world%rank == 0) then
    write(io_out,*) 'SETUP CONDITION>  =================================>'
    write(io_out,*) ' '
  end if

  call setup(world, main)
  call MPI_BARRIER(world%comm, ierr)

  t(4) = mpi_wtime()
  dt(3) = t(4) - t(3)


  ! Main loop
  !

  if (world%rank == 0) then
    write(io_out,*) 'MAIN CALCULATION>  ================================>'
    write(io_out,*) ' '
  end if

  call driver(world, main)

  t(5) = mpi_wtime()
  dt(4) = t(5) - t(4)


  ! Finalize MPI
  !

  if (world%rank == 0) then
    write(io_out,*) 'FINALIZE>  ========================================>'
    write(io_out,*) ' '
  end if

  call MPI_BARRIER(world%comm, ierr)

  if (world%rank == 0) then
    write(io_out,'(A)')      ' ============= MPI WTIMER ============='
    write(io_out,'(A,F9.3,A)') ' SETTING AND READ INPUT :', dt(1), ' sec.'
    write(io_out,'(A,F9.3,A)') ' BCASTING INITIAL DATA  :', dt(2), ' sec.'
    write(io_out,'(A,F9.3,A)') ' READ CONDITION DATA    :', dt(3), ' sec.'
    write(io_out,'(A,F9.3,A)') ' MAIN LOOP              :', dt(4), ' sec.'
    write(io_out,'(A,F9.3,A)') ' SYSTEM TIME            :', &
                                          dt(1)+dt(2)+dt(3)+dt(4), ' sec.'
    write(io_out,'(A)')      ' ======================================'
    write(io_out,'(A)')      ' '
  end if

  call MPI_FINALIZE (ierr)

  stop

end program REIN
