!-------------------------------------------------------------------------------
!    This source code is part of
!    REIN (Replica-Exchange INterface program) version 1.0
!
!    Written by Naoyuki Miyashita and Yuji Sugita
!    Copyright (c) 2012, RIKEN, Japan
!
!    This program is free software; you can redistribute it and/or
!    modify it under the terms of the GNU General Public License
!    as published by the Free Software Foundation, either 
!    version 3 of the License, or (at your option) any later version.
!
!    This program is distributed in the hope that it will be useful,
!    but WITHOUT ANY WARRANTY; without even the implied 
!    warranty of MERCHANTABILITY or FITNESS FOR 
!    A PARTICULAR PURPOSE.  
!    See the GNU General Public License for more details.
!
!    You should have received a copy of the GNU General 
!    Public License along with REIN.  
!    If not, see <http://www.gnu.org/licenses/>.
!-------------------------------------------------------------------------------

!--------1---------2---------3---------4---------5---------6---------7---------8
!
!  Module   bb_batch_mod
!> @brief   build batch script file
!! @authors Norio Takase (NT), Naoyuki Miyashita (NM)
!! @date    2012/05/31
! 
!--------1---------2---------3---------4---------5---------6---------7---------8

module bb_batch_mod

  use bb_reininp_mod
  use bb_ctrl_mod
  use replace_string_mod
  use fileio_control_mod
  use fileio_mod
  use messages_mod

  implicit none
  private

  ! parameters
  integer, parameter :: StageIn     = 1
  integer, parameter :: StageInDir  = 2
  integer, parameter :: StageOut    = 3
  integer, parameter :: StageOutDir = 4

  ! subroutines
  public  :: batch
  private :: batch_main
  private :: batch_stagein
  private :: batch_stageout
  private :: batch_staging

contains

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine  batch
  !> @brief      batch batch script file
  !! @authors    NM, NT
  !! @param[inout]  ctrl_data   control parameters in the program
  !! @param[in]  rein_inp    information of REIN input
  !! @date       2012/05/31
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine batch(ctrl_data, rein_inp)

    ! formal arguments
    type(s_ctrl_data), intent(inout) :: ctrl_data
    type(s_rein_inp),  intent(in)    :: rein_inp

    ! local variables
    integer            :: bat_out
    character(100)     :: job_scheduler, bat_name


    bat_name = ctrl_data%batch_info%batchfile_name 

    write(MsgOut,'(A,A)') 'Batch> Batch queue file: ', TRIM(bat_name)

    call open_file(bat_out, bat_name, IOFileOutputReplace)

    call batch_main(ctrl_data, rein_inp, bat_out)
    
    call close_file(bat_out)


    write(MsgOut,'(A)') ' '

    return

  end subroutine batch


  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine  batch_main
  !> @brief      build batch script file
  !! @authors    NM, NT
  !! @param[inout]  ctrl_data   control parameters in the program
  !! @param[in]  rein_inp    information of REIN input
  !! @param[in]  bat_out     file unit no for batch file
  !! @date       2012/05/31 
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine batch_main(ctrl_data, rein_inp, bat_out)

    ! formal arguments
    type(s_ctrl_data), intent(inout) :: ctrl_data
    type(s_rein_inp),  intent(in)    :: rein_inp
    integer,           intent(in)    :: bat_out

    ! local variables
    integer            :: tmp_in, kw_index, i, num
    character(200)     :: line, job_scheduler
    character(50)      :: keyword, val
    integer            :: num_procs_rein

    ! pre processing

    if ( TRIM( ctrl_data%batch_info%direct_describe ) .eq. 'no' ) then
      if( TRIM( ctrl_data%main_info%openmp ) .eq. "yes")              &
                ctrl_data%batch_info%num_mpi_static_node_shape        &
                =  ctrl_data%batch_info%num_mpiproc_rein
      if( TRIM( ctrl_data%main_info%hybrid_md ) .eq. "no" .and.  &
          TRIM( ctrl_data%main_info%openmp )    .eq. "no"     )  &
          ctrl_data%batch_info%num_mpi_static_node_shape         &
          =  int( ( ctrl_data%batch_info%num_mpiproc_rein - 1 )  &
                / ctrl_data%batch_info%num_core_cpu      )       &
             + 1
      if( TRIM( ctrl_data%main_info%hybrid_md ) .eq. "yes" .and. &
          TRIM( ctrl_data%main_info%openmp )    .eq. "no"     )  &
          ctrl_data%batch_info%num_mpi_static_node_shape         &
          =  int( ctrl_data%batch_info%num_mpiproc_rein )  
    end if


    ! open the job scheduler template file
    !
    job_scheduler = ctrl_data%batch_info%job_scheduler
    call tolower(job_scheduler)


    call open_file(tmp_in, &
         TRIM(ctrl_data%main_info%template_dir)//'/t_batch_'// &
         TRIM(job_scheduler), IOFileInput)

    do while (.true.)
      read(tmp_in,'(A)',end=10) line
      
      do while (check_replace(line, kw_index, keyword))
        
        select case(keyword)
          
        case ('BATCH_BUILDER_RESTART')
          call replace(line, kw_index, keyword,         &
                       ctrl_data%main_info%batch_builder_restart)

        case ('JOBNAME')
          call replace(line, kw_index, keyword, &
                       ctrl_data%main_info%jobname)

        case ('CPU_TIME')
          call replace(line, kw_index, keyword, &
                       ctrl_data%batch_info%cpu_time)

        case ('MACHINE_FILE')
          call replace(line, kw_index, keyword, &
                       ctrl_data%batch_info%machine_file)

        case ('MPI_STATIC_NODE_SHAPE')

          if ( TRIM( ctrl_data%batch_info%direct_describe ) .eq. 'yes' ) then
            if (ctrl_data%batch_info%mpi_static_node_shape == '') &
              call error_msg('Batch_Main> mpi_static_node_shape is empty')
            call replace(line, kw_index, keyword, &
                         ctrl_data%batch_info%mpi_static_node_shape)
          else
            write(val,'(I0)') ctrl_data%batch_info%num_mpi_static_node_shape
            call replace(line, kw_index, keyword, val)
          end if

        case ('MPI_STATIC_NODE_PROC')

          if ( TRIM( ctrl_data%batch_info%direct_describe ) .eq. 'yes' ) then
            if (ctrl_data%batch_info%mpi_static_node_proc == '') &
              call error_msg('Batch_Main> mpi_static_node_proc is empty')
            call replace(line, kw_index, keyword, &
                         ctrl_data%batch_info%mpi_static_node_proc)
          else

            if ( TRIM (ctrl_data%main_info%hybrid_md) .eq. 'yes' ) then
              ctrl_data%batch_info%num_mpi_static_node_proc        = &
                  ctrl_data%batch_info%num_mpi_static_node_shape !    &
              !  * ctrl_data%batch_info%num_core_cpu                  &
              !  / ctrl_data%batch_info%num_threads_md
            else
              ctrl_data%batch_info%num_mpi_static_node_proc        = &
                  ctrl_data%batch_info%num_mpi_static_node_shape     &
                * ctrl_data%batch_info%num_core_cpu 
            end if

            write(val,'(I0)') ctrl_data%batch_info%num_mpi_static_node_proc
            call replace(line, kw_index, keyword, val)
          end if

        case ('NODE')

          if ( TRIM( ctrl_data%batch_info%direct_describe ) .eq. 'yes' ) then
            if (ctrl_data%batch_info%node == '') &
              call error_msg('Batch_Main> node is empty')
            call replace(line, kw_index, keyword, &
                         ctrl_data%batch_info%node)
          else
            num = ctrl_data%batch_info%num_mpiproc_md            &
                * ctrl_data%batch_info%num_threads_md            &
                * rein_inp%number_of_replicas                    &
                / ctrl_data%batch_info%num_core_cpu 

            if ( TRIM( ctrl_data%main_info%spawn ) .eq. 'yes' ) then
              num = num + ctrl_data%batch_info%num_mpi_static_node_shape
            else if ( TRIM( ctrl_data%batch_info%job_scheduler ) .eq. 'GE' ) then
                num = num * ctrl_data%batch_info%num_core_cpu
            else if ( TRIM( ctrl_data%batch_info%job_scheduler ) .eq. 'RICCTAR' ) then
                if( ctrl_data%batch_info%num_threads_md > 1 )     &
                 num = rein_inp%number_of_replicas
            else if ( TRIM( ctrl_data%batch_info%job_scheduler ) .eq. 'RICC' ) then
                if( ctrl_data%batch_info%num_threads_md > 1 )     &
                 num = rein_inp%number_of_replicas
            end if

            write(val,'(I0)') num
            call replace(line, kw_index, keyword, val)
          end if

        case ('NUM_MPIPROC_REIN')
            if ( ( TRIM( ctrl_data%batch_info%job_scheduler ) .eq. 'K' ) &
               .and.    &
            ( TRIM( ctrl_data%main_info%hybrid_md ) .eq. "no" ) ) then
              num_procs_rein = ctrl_data%batch_info%num_mpiproc_rein ! &
                     !          * ctrl_data%batch_info%num_core_cpu 
               if( num_procs_rein > rein_inp%number_of_replicas ) then
                 num_procs_rein = rein_inp%number_of_replicas
               end if
            else
              num_procs_rein = ctrl_data%batch_info%num_mpiproc_rein
               if( num_procs_rein > rein_inp%number_of_replicas ) then
                 num_procs_rein = rein_inp%number_of_replicas
               end if
            end if
            !write(val,'(I0)') ctrl_data%batch_info%num_mpiproc_rein
            write(val,'(I0)') num_procs_rein
            call replace(line, kw_index, keyword, val)

        case ('NUM_REPLICAS')
          write(val,'(I0)') rein_inp%number_of_replicas
          call replace(line, kw_index, keyword, val)

        case ('NUMBER_OF_THREADS')
          write(val,'(I0)') ctrl_data%batch_info%num_threads_md
          if( TRIM ( ctrl_data%main_info%openmp )    .eq. 'no' .and.  &
              TRIM ( ctrl_data%main_info%hybrid_md ) .eq. 'no'     )  &
            val = '1'
          if( TRIM ( ctrl_data%main_info%openmp )    .eq. 'no' .and.  &
              TRIM ( ctrl_data%main_info%hybrid_md ) .eq. 'yes'     ) then 
            write(val,'(I0)') ctrl_data%batch_info%num_core_cpu
          end if
          if( TRIM ( ctrl_data%main_info%openmp )    .eq. 'yes' .and.  &
              TRIM ( ctrl_data%main_info%hybrid_md ) .eq. 'yes'     ) then
            write(val,'(I0)') ctrl_data%batch_info%num_core_cpu
          end if
          if( TRIM ( ctrl_data%main_info%spawn  ) .eq. 'no' ) then
            write(val,'(I0)') ctrl_data%batch_info%num_threads_md
          end if

          call replace(line, kw_index, keyword, val)

        case ('PARALLEL')
          call replace(line, kw_index, keyword, &
                       ctrl_data%main_info%parallel)

        case ('PROCESS')

          if ( TRIM( ctrl_data%batch_info%direct_describe ) .eq. 'yes' ) then
            if (ctrl_data%batch_info%process == '') &
              call error_msg('Batch_Main> process is empty')
            call replace(line, kw_index, keyword, &
                         ctrl_data%batch_info%process)
          else
            num = ctrl_data%batch_info%num_mpiproc_md            &
                * ctrl_data%batch_info%num_threads_md            &
                * rein_inp%number_of_replicas

            if ( TRIM( ctrl_data%main_info%spawn ) .eq. 'yes' ) then
              num = num + ctrl_data%batch_info%num_mpi_static_node_shape
            else if ( TRIM( ctrl_data%batch_info%job_scheduler ) .eq. 'GE' ) then
                num = num
            end if

            write(val,'(I0)') num
            call replace(line, kw_index, keyword, val)
          end if

        case ('QUEUE_NAME')
          call replace(line, kw_index, keyword, &
                       ctrl_data%batch_info%queue_name)

        case ('REIN_INPUT')
          call replace(line, kw_index, keyword, &
                       ctrl_data%main_info%rein_input)

        case ('STGIN')
          call batch_stagein(ctrl_data, rein_inp, bat_out)
          line = '#'

        case ('STGOUT')
          call batch_stageout(ctrl_data, rein_inp, bat_out)
          line = '#'

        case ('STAGE_OUT_DIR')
          call replace(line, kw_index, keyword, &
               ctrl_data%batch_info%stage_out_dir)

        case ('ATMARK')
          call replace(line, kw_index, keyword, '?')

        case default
          write(MsgOut,*) "Batch_main> Unknown template keyword : "// &
                          TRIM(keyword)
          exit

        end select

      end do

      do i = 1, LEN_TRIM(line)
        if (line(i:i) == '?') line(i:i) = '@'
      end do

      write(bat_out,'(A)') TRIM(line)

    end do

10  call close_file(tmp_in)

    return

  end subroutine batch_main


  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine  batch_stagein
  !> @brief      write out stage in files
  !! @authors    NT
  !! @param[in]  ctrl_data control parameters in the program
  !! @param[in]  rein_inp  information of REIN input
  !! @param[in]  bat_out   file unit no for batch file
  !! @date       2012/10/10 
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine batch_stagein(ctrl_data, rein_inp, bat_out)

    ! formal arguments
    type(s_ctrl_data), intent(in)    :: ctrl_data
    type(s_rein_inp),  intent(in)    :: rein_inp
    integer,           intent(in)    :: bat_out

    ! local variables
    integer            :: i, ctrl_in
    character(100)     :: job_scheduler, program_name, filename


    ! open template file
    !

    job_scheduler = ctrl_data%batch_info%job_scheduler
    program_name  = rein_inp%program_name
    call tolower(job_scheduler)
    call tolower(program_name)

    filename = TRIM(ctrl_data%main_info%template_dir)//'/t_staging_'// &
               TRIM(job_scheduler)//'_'//TRIM(program_name)

    call open_ctrlfile(filename, ctrl_in)


    ! write out stage-in files
    !

    do i = 1, rein_inp%number_of_replicas
      call batch_staging(ctrl_data, rein_inp, StageIn,    i, bat_out, ctrl_in)
      call batch_staging(ctrl_data, rein_inp, StageInDir, i, bat_out, ctrl_in)
    end do

    call batch_staging(ctrl_data, rein_inp, StageIn,    0, bat_out, ctrl_in)
    call batch_staging(ctrl_data, rein_inp, StageInDir, 0, bat_out, ctrl_in)


    ! close template files
    !

    call close_ctrlfile(ctrl_in)

    return

  end subroutine batch_stagein


  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine  batch_stageout
  !> @brief      write out stage out files
  !! @authors    NT
  !! @param[in]  ctrl_data control parameters in the program
  !! @param[in]  rein_inp  information of REIN input
  !! @param[in]  bat_out   file unit no for batch file
  !! @date       2012/10/10 
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine batch_stageout(ctrl_data, rein_inp, bat_out)

    ! formal arguments
    type(s_ctrl_data), intent(in)    :: ctrl_data
    type(s_rein_inp),  intent(in)    :: rein_inp
    integer,           intent(in)    :: bat_out

    ! local variables
    integer            :: i, ctrl_in
    character(100)     :: job_scheduler, program_name


    ! open template file
    !

    job_scheduler = ctrl_data%batch_info%job_scheduler
    program_name  = rein_inp%program_name
    call tolower(job_scheduler)
    call tolower(program_name)

    call open_ctrlfile( &
         TRIM(ctrl_data%main_info%template_dir)//'/t_staging_'// &
         TRIM(job_scheduler)//'_'// &
         TRIM(program_name), ctrl_in)


    ! write out stage-out files
    !

    call batch_staging(ctrl_data, rein_inp, StageOutDir, 0, bat_out, ctrl_in)
    call batch_staging(ctrl_data, rein_inp, StageOut,    0, bat_out, ctrl_in)

    do i = 1, rein_inp%number_of_replicas
      call batch_staging(ctrl_data, rein_inp, StageOutDir, i, bat_out, ctrl_in)
      call batch_staging(ctrl_data, rein_inp, StageOut,    i, bat_out, ctrl_in)
    end do


    ! close template files
    !

    call close_ctrlfile(ctrl_in)

    return

  end subroutine batch_stageout


  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine  batch_staging
  !> @brief      write out stage in/out replica files
  !! @authors    NM, NT
  !! @param[in]  ctrl_data control parameters in the program
  !! @param[in]  rein_inp  information of REIN input
  !! @param[in]  staging   mode for staging
  !! @param[in]  rep_no    replica number
  !! @param[in]  bat_out   file unit no for batch file
  !! @date       2012/06/02 
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine batch_staging(ctrl_data, rein_inp, staging, rep_no, &
                           bat_out, ctrl_in)

    ! formal arguments
    type(s_ctrl_data), intent(in)    :: ctrl_data
    type(s_rein_inp),  intent(in)    :: rein_inp
    integer,           intent(in)    :: staging
    integer,           intent(in)    :: rep_no
    integer,           intent(in)    :: bat_out
    integer,           intent(in)    :: ctrl_in

    ! local variables
    integer            :: system
    integer            :: i, j, kw_index
    character(200)     :: val, val0, val1, val_src
    character(50)      :: keyword, prefix, source, quotation
    character(20)      :: key, section


    select case (staging)

    case (StageIn)
      section = 'STAGEIN'
    case (StageInDir)
      section = 'STAGEINDIR'
    case (StageOut)
      section = 'STAGEOUT'
    case (StageOutDir)
      section = 'STAGEOUTDIR'

    end select

    prefix = ''
    call read_ctrlfile_string(ctrl_in, section, 'Prefix', prefix)

    source = ''
    call read_ctrlfile_string(ctrl_in, section, 'Source', source)

    quotation = ''
    call read_ctrlfile_string(ctrl_in, section, 'Quotation', quotation)


    i = 0
    do while (.true.)
      i = i + 1

      if (rep_no /= 0) then
        write(key,'(A7,I3.3)') 'Replica', i
      else
        write(key,'(A4,I3.3)') 'File', i
      end if

      val = ''
      call read_ctrlfile_string(ctrl_in, section, key, val)
      if (LEN_TRIM(val) == 0) &
         exit

      do while(check_replace(val, kw_index, keyword))

        select case(keyword)
          
        case ('PARALLEL_EXEC')
          call replace(val, kw_index, keyword,         &
            rein_inp%parallel_exec)

        case ('BATCHBUILDER_INPUT')
          call replace(val, kw_index, keyword,         &
            ctrl_data%main_info%batch_builder_input )

        case ('BATCHBUILDER_EXE')
          call replace(val, kw_index, keyword,         &
            ctrl_data%main_info%batch_builder_exe )

        case ('FIRST_STEP')
          write(val0,'(I0)') rein_inp%first_step-1
          call replace(val, kw_index, keyword, val0)

        case ('INPUT_FILE1')
          call replace(val, kw_index, keyword, &
               rein_inp%inp_files(1))

        case ('INPUT_FILE2')
          call replace(val, kw_index, keyword, &
               rein_inp%inp_files(2))

        case ('INPUT_FILE3')
          call replace(val, kw_index, keyword, &
               rein_inp%inp_files(3))

        case ('INPUT_FILE4')
          call replace(val, kw_index, keyword, &
               rein_inp%inp_files(4))

        case ('INPUT_FILE5')
          call replace(val, kw_index, keyword, &
               rein_inp%inp_files(5))

        case ('INPUT_FILE6')
          call replace(val, kw_index, keyword, &
               rein_inp%inp_files(6))

        case ('OUTPUT_FILE1')
          call replace(val, kw_index, keyword, &
               rein_inp%out_files(1))

        case ('OUTPUT_FILE2')
          call replace(val, kw_index, keyword, &
               rein_inp%out_files(2))

        case ('OUTPUT_FILE3')
          call replace(val, kw_index, keyword, &
               rein_inp%out_files(3))

        case ('REIN_INPUT')
          call replace(val, kw_index, keyword, &
               ctrl_data%main_info%rein_input)

        case ('REP_NAME')
          call replace(val, kw_index, keyword, rein_inp%io_file_name)

        case ('REP_NO')
          write(val0,'(I0)') rep_no
          call replace(val, kw_index, keyword, val0)

        case ('REP_NO-1')
          write(val0,'(I0)') rep_no - 1
          call replace(val, kw_index, keyword, val0)

        case ('STAGE_OUT_DIR')
          call replace(val, kw_index, keyword, &
               ctrl_data%batch_info%stage_out_dir)

        case ('ATMARK')
          call replace(val, kw_index, keyword, '?')

        case default
          write(MsgOut,*) "Batch_staging> Unknown template keyword : "//&
               TRIM(keyword)
          exit

        end select

      end do

      do j = 1, LEN_TRIM(val)
        if (val(j:j) == '?') val(j:j) = '@'
      end do

      if (staging == StageIn .or. staging == StageInDir) then
        do j = 1, LEN_TRIM(val)
          if (val(j:j) == '/') val(j:j) = '?'
        end do
        val0 = ''
        val1 = ''
        read(val,*,end=100) val0, val1
100     do j = 1, LEN_TRIM(val0)
          if (val0(j:j) == '?') val0(j:j) = '/'
        end do
        do j = 1, LEN_TRIM(val1)
          if (val1(j:j) == '?') val1(j:j) = '/'
        end do
        do j = 1, LEN_TRIM(val)
          if (val(j:j) == '?') val(j:j) = '/'
        end do

        if (source == 'left' .or. val1 == '') then
          val_src = val0
        else
          val_src = val1
        end if

        if (system('test -e '//TRIM(val_src)) /= 0) then
          write(MsgOut,*) &
          'Batch_staging> WARNING: file not found. skip.. ['//TRIM(val_src)//']'
          cycle
        end if
      end if

      write(bat_out,'(A,A,A,A,A,A)') &
           '#', TRIM(prefix), ' ', TRIM(quotation), TRIM(val), TRIM(quotation)
    end do

    return

  end subroutine batch_staging

end module bb_batch_mod
